# Standard imports
import argparse
import time
from pathlib import Path
from math import ceil
from tqdm.auto import tqdm

# Medical imaging imports
import nibabel as nib
from radifox.utils.resize.affine import update_affine

# ML imports
import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from .models.wdsr import WDSR

# Utils folder
from .utils.train_set import TrainSet
from .utils.timer import timer_context
from .utils.parse_image_file import (
    lr_axis_to_z,
    z_axis_to_lr_axis,
    inv_normalize,
)
from .utils.misc_utils import LossProgBar
from .utils.patch_ops import calc_dims_to_crop, calc_dims_to_pad
from .utils.rotate import rotate_vol_2d

torch.backends.cudnn.benchmark = True


def set_random_seed(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def apply_to_vol(model, image, batch_size):
    result = []
    for st in tqdm(range(0, image.shape[0], batch_size)):
        en = st + batch_size
        batch = image[st:en]
        with torch.inference_mode():
            with torch.amp.autocast("cuda"):
                sr = model(batch).detach().cpu()
        result.append(sr)
    result = torch.cat(result, dim=0)
    return result


set_random_seed(0)
text_div = "=" * 10


def run_eclare(
    image,
    slice_separation,
    scales,
    lr_axis,
    header,
    affine,
    orig_min,
    orig_max,
    blur_kernel,
    dataset,
    device,
    batch_size,
    n_patches=None,
):

    eclare_st = time.time()

    learning_rate = 1e-4

    # model setup
    model = WDSR(
        n_resblocks=16,
        num_channels=256,
        scale=slice_separation,
    ).to(device)

    p = 8
    p2 = round(slice_separation * max(model.calc_out_patch_size([p, p])))
    dataset.init_eclare((p2, p), blur_kernel, slice_separation, n_patches)

    n_steps = int(ceil(dataset.n_patches / batch_size))

    opt = torch.optim.Adam(model.parameters(), lr=learning_rate)
    scheduler = torch.optim.lr_scheduler.OneCycleLR(
        optimizer=opt,
        max_lr=learning_rate,
        total_steps=n_steps + 1,
        cycle_momentum=True,
    )
    opt.step()  # necessary for the LR scheduler
    scaler = torch.amp.GradScaler("cuda")
    loss_obj = torch.nn.MSELoss().to(device)

    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,  # Dataset automatically shuffles
        pin_memory=True,
        num_workers=min(8, torch.get_num_threads()),
    )

    # ===============================
    # ============ TRAIN ============
    # ===============================
    print(f"\n{text_div} Running ECLARE {text_div}")
    train_st = time.time()

    loss_names = ["loss"]
    pbar = LossProgBar(dataset.n_patches, batch_size, loss_names)

    for t, (patches_lr, patches_hr) in enumerate(data_loader):
        patches_hr_device = patches_hr.to(device)
        patches_lr_device = patches_lr.to(device)

        with torch.amp.autocast("cuda"):
            patches_hr_hat = model(patches_lr_device)
            loss = loss_obj(patches_hr_hat, patches_hr_device)

        # gradient updates
        opt.zero_grad()
        scaler.scale(loss).backward()
        scaler.step(opt)
        scaler.update()
        scheduler.step()

        # Update progress bar
        pbar.update({"loss": loss})

    pbar.close()

    # ===============================
    # ========== INFERENCE ==========
    # ===============================

    # Uses variables instantiated way earlier
    image = lr_axis_to_z(image, lr_axis)
    image = torch.from_numpy(image)

    # ===== PREDICT =====
    angles = [0, 90]

    model_preds = []

    for i, angle in enumerate(angles):
        context_str = f"Super-resolving at {angle} degrees"
        with timer_context(context_str, verbose=False):
            # Rotate in-plane. Image starts as (hr_axis, hr_axis, lr_axis)
            image_rot = rotate_vol_2d(image.to(device), angle)
            # Ensure the LR axis is s.t. (hr_axis, C, lr_axis, hr_axis)
            image_rot = image_rot.permute(0, 2, 1).unsqueeze(1)
            # Run model
            rot_result = apply_to_vol(model, image_rot, 1)
            # Return to (hr_axis, hr_axis, lr_axis)
            result = rot_result.squeeze(1).permute(0, 2, 1)
            # Rotate back and collect
            model_preds.append(rotate_vol_2d(result, -angle))

    # ===== FINALIZE =====
    final_out = torch.mean(torch.stack(model_preds), dim=0)
    final_out = final_out.detach().cpu().numpy().astype(np.float32)
    final_out = inv_normalize(final_out, orig_min, orig_max, a=0, b=1)

    # Reorient to original orientation
    final_out = z_axis_to_lr_axis(final_out, lr_axis)

    print(f"{text_div} End prediction {text_div}")

    eclare_en = time.time()

    return final_out, model.state_dict(), eclare_en - eclare_st



"""
Adapted from https://github.com/JiahuiYu/wdsr_ntire2018/blob/master/wdsr_b.py
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from radifox.utils.resize.pytorch import resize
from math import ceil
import numpy as np


def pixel_shuffle(x, scale):
    """https://gist.github.com/davidaknowles/6e95a643adaf3960d1648a6b369e9d0b"""
    num_batches, num_channels, nx, ny = x.shape
    num_channels = num_channels // scale
    out = x.contiguous().view(num_batches, num_channels, scale, nx, ny)
    out = out.permute(0, 1, 3, 2, 4).contiguous()
    out = out.view(num_batches, num_channels, nx * scale, ny)
    return out


class Upsample(nn.Module):
    def __init__(self, num_channels, scale, kernel_size):
        super().__init__()
        padding = (kernel_size - 1) // 2
        self.scale = scale
        self.conv0 = nn.Conv2d(num_channels, scale, kernel_size, padding=padding)

    def forward(self, x):
        out = self.conv0(x)
        out = pixel_shuffle(out, self.scale)
        return out


class Block(nn.Module):
    def __init__(self, n_feats, act=nn.ReLU(True), res_scale=1):
        super(Block, self).__init__()
        self.res_scale = res_scale
        expand = 4
        linear = 0.8
        self.body = nn.Sequential(
            *[
                nn.Conv2d(n_feats, n_feats * expand, 1, padding=0),
                act,
                nn.Conv2d(n_feats * expand, int(n_feats * linear), 1, padding=0),
                nn.Conv2d(int(n_feats * linear), n_feats, 3, padding=1),
            ]
        )

    def forward(self, x):
        return x + self.body(x) * self.res_scale


class WDSR(nn.Module):
    def __init__(self, n_resblocks, num_channels, scale, order=3):
        super().__init__()
        self.scale = scale
        self.order = order

        kernel_size = 3
        padding = (kernel_size - 1) // 2
        act = nn.ReLU(True)

        self.head = nn.Conv2d(1, num_channels, kernel_size, padding=padding)

        self.body = nn.Sequential(
            *[Block(num_channels, act=act, res_scale=1) for _ in range(n_resblocks)]
        )

        self.tail = Upsample(num_channels, ceil(self.scale), kernel_size)

    def calc_out_patch_size(self, input_patch_size):
        x = torch.rand([1, 1] + input_patch_size).float()
        x = x.to(next(self.parameters()).device)
        out, *_ = self(x)
        patch_size = list(out.shape[2:])

        return patch_size

    def forward(self, x):
        # Interpolate the input to create a skip
        s = resize(x, (1 / self.scale, 1), order=self.order)

        # Process in LR space
        x = self.head(x)
        x = self.body(x)
        # Pixel shuffle to ceil(scale)
        x = self.tail(x)
        # Interpolate down to target shape
        x = resize(x, (ceil(self.scale) / self.scale, 1), order=self.order)

        # Track the residual
        r = x.clone()

        # Add the residual
        x = r + s

        # Clip to the range [0, 1]
        x = x.clip(0, 1)
        return x



import time
from contextlib import contextmanager


@contextmanager
def timer_context(label, verbose=True):
    if verbose:
        print(label)
    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        elapsed_time = end_time - start_time
        if verbose:  # Print elapsed time only if verbose is True
            print(f"\tElapsed time: {elapsed_time:.4f}s")



import numpy as np
import itertools


def augment_3d_image(image):
    """
    As augmentation, we hold all combinations of horizontal and vertical flips
    as well as transposition for the in-plane of a 3D image.

    This function is written by GPT-4 Code Interpreter and edited by a human.
    """
    assert len(image.shape) == 3, "Input must be a 3D numpy array."

    augmented_images = []
    for flip_h, flip_v, do_transpose in itertools.product([False, True], repeat=3):
        # Copy the original image to avoid modifying it
        augmented_image = np.copy(image)

        # Flip horizontally if flip_h is True
        if flip_h:
            augmented_image = np.fliplr(augmented_image)

        # Flip vertically if flip_v is True
        if flip_v:
            augmented_image = np.flipud(augmented_image)

        # Transpose in the AxA plane if do_transpose is True
        if do_transpose:
            augmented_image = np.transpose(augmented_image, axes=(1, 0, 2))

        augmented_images.append(np.copy(augmented_image))

    return augmented_images



# Standard imports
import argparse
import time
from pathlib import Path
from math import ceil
import matplotlib
import matplotlib.pyplot as plt
from tqdm.auto import tqdm

# Medical imaging imports
from radifox.utils.resize.pytorch import resize as resize_pt

# ML imports
import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from .models.espreso_g import G
from .models.espreso_d import D
from .models.losses import *


# Utils folder
from .utils.train_set import TrainSet
from .utils.fwhm import calc_fwhm


def set_random_seed(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


set_random_seed(0)
matplotlib.use("Agg")  # non-interactive plotting
text_div = "=" * 10


def run_espreso(
    slice_separation,
    dataset,
    device,
    espreso_psf_fpath,
    espreso_psf_plot_fpath,
):

    espreso_st = time.time()
    print(f"{text_div} Running ESPRESO {text_div}")

    lr_patch_size = [48, 48]
    batch_size = 128
    lr = 1e-4

    # ESPRESO starts with P x 16 patches as its goal. The "generator" produces a blur kernel
    # which convolves along the first axis without padding, and then downsamples along that
    # axis. Therefore P must be 16 + 2\floor{L/2} x k, where L is the kernel length and k is
    # the slice separation.
    P = 16 * slice_separation + 2 * 10
    patch_size = (int(round(P)), 16)

    # warmup data loader
    n_patches = batch_size * 100
    n_steps = int(ceil(n_patches / batch_size))

    dataset.init_espreso(patch_size, n_patches)
    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,  # Dataset automatically shuffles
        pin_memory=False,
        num_workers=min(8, torch.get_num_threads()),
    )

    g = G().to(device)
    opt_g = torch.optim.AdamW(g.parameters(), lr=lr)
    scaler_g = torch.amp.GradScaler("cuda")

    # Warm up to RF-Pulse-SLR of the slice separation
    ref_psf = torch.zeros(1, 1, 21, 1).to(device)
    ref_psf[:, :, 21 // 2, :] = 1

    loss_fn = torch.nn.MSELoss().to(device)

    for x_batch in tqdm(data_loader):

        x = x_batch.to(device)

        opt_g.zero_grad()

        with torch.amp.autocast("cuda"):
            est_psf = g()

            est = F.conv2d(x, est_psf)
            ref = F.conv2d(x, ref_psf)

            loss = loss_fn(est, ref)

        scaler_g.scale(loss).backward()

        scaler_g.step(opt_g)
        scaler_g.update()

    est_psf_warmup = g().detach().cpu().numpy().squeeze()

    ### Adversarial training

    d_patch_size = (10, 16)

    adv_loss_fn = GANLoss().to(device)
    peak_loss_fn = PeakLoss().to(device)
    bound_loss_fn = BoundaryLoss(kernel_length=21).to(device)

    # now adversarial loss.
    # "real" means that the horizontal direction is from the through-plane axis
    # "fake" means that the horizontal direction is NOT from the through-plane axis

    # adversarial training uses more training patches
    dataset.n_patches = 128 * 2 * 2000
    n_steps = int(ceil(dataset.n_patches / 128 * 2))

    # re-initialize the data_laoder with new batch size
    data_loader = DataLoader(
        dataset,
        batch_size=int(data_loader.batch_size * 2),
        shuffle=False,  # Dataset automatically shuffles
        pin_memory=True,
        num_workers=8,
    )

    # Initialize discriminator
    d = D().to(device)
    opt_d = torch.optim.AdamW(d.parameters(), lr=lr)

    # Set up LR schedulers
    sch_g = torch.optim.lr_scheduler.OneCycleLR(
        optimizer=opt_g,
        max_lr=lr,
        total_steps=n_steps + 1,
        cycle_momentum=True,
    )
    sch_d = torch.optim.lr_scheduler.OneCycleLR(
        optimizer=opt_d,
        max_lr=lr,
        total_steps=n_steps + 1,
        cycle_momentum=True,
    )

    # Step optimizers for schedulers
    opt_g.step()
    opt_d.step()

    # set up scalers
    scaler_g = torch.amp.GradScaler("cuda")
    scaler_d = torch.amp.GradScaler("cuda")

    for x_batch in tqdm(data_loader):

        x = x_batch.to(device)

        # ===== Train generator ===== #
        # freeze discriminator
        for p in d.parameters():
            p.requires_grad = False

        opt_g.zero_grad()

        with torch.amp.autocast("cuda"):

            est_psf = g()

            y = F.conv2d(x, est_psf)
            y = resize_pt(y, (slice_separation, 1), order=3)

            real_horiz = y[: batch_size // 2].permute(0, 1, 3, 2)
            fake_horiz = y[batch_size // 2 :]

            N = real_horiz.shape[0]
            g_loss = adv_loss_fn(d(fake_horiz), is_real=True)
            p_loss = peak_loss_fn(est_psf)
            b_loss = bound_loss_fn(est_psf)

            loss = g_loss + 0.5 * p_loss + 10 * b_loss

        scaler_g.scale(loss).backward()
        scaler_g.step(opt_g)
        scaler_g.update()

        sch_g.step()

        # ===== Train discriminator ===== #
        # unfreeze the discriminator
        for p in d.parameters():
            p.requires_grad = True

        opt_d.zero_grad()

        with torch.amp.autocast("cuda"):
            # Measure discriminator's ability to classify real from generated samples
            if np.random.rand() < 0.5:
                d_loss = adv_loss_fn(d(real_horiz.detach()), is_real=True)
            else:
                d_loss = adv_loss_fn(d(fake_horiz.detach()), is_real=False)

        scaler_d.scale(d_loss).backward()
        scaler_d.step(opt_d)
        scaler_d.update()

        sch_d.step()

    espreso_psf = g().detach().cpu().numpy().squeeze()

    # save PSF to disk
    with open(espreso_psf_fpath, "wb") as fpointer:
        np.save(fpointer, espreso_psf)

    # save plot of PSF to disk
    fwhm = calc_fwhm(espreso_psf)[0]
    plt.plot(espreso_psf, color="blue")

    title = f"Calculated FWHM: {fwhm:.2f}"
    plt.title(title)
    plt.savefig(espreso_psf_plot_fpath)

    espreso_en = time.time()

    return g, espreso_en - espreso_st



import torch
from tqdm.auto import tqdm
from functools import wraps


def parse_device(gpu_id):
    if torch.cuda.is_available() and gpu_id >= 0:
        device = torch.device(f"cuda:{gpu_id}")
    else:
        print("GPU index not provided or no GPU support currently available.")
        print("!!! Running on CPU !!!")
        device = torch.device("cpu")
    return device


class LossProgBar:
    def __init__(self, total, update_amount, loss_names, precision=4):
        self.total = total
        self.update_amount = update_amount
        self.precision = precision
        self.pbar = tqdm(total=self.total)
        self.pbar_dict = {name: torch.finfo(torch.float32).max for name in loss_names}

    def __enter__(self):
        self.pbar = tqdm(total=self.total)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.pbar.close()

    def close(self):
        self.pbar.close()

    def update(self, loss_dict):
        for k, v in loss_dict.items():
            self.pbar_dict[k] = v.detach().cpu().numpy().item()
        self.pbar.set_postfix(
            {k: f"{v:.{self.precision}f}" for k, v in self.pbar_dict.items() if v != 0}
        )
        self.pbar.update(self.update_amount)



import numpy as np
from scipy.ndimage import gaussian_filter
from math import ceil, floor


def get_ilb(m, s, a, b):
    return round(m * s) / a - m - 1 / (2 * a)


def get_plb(m, s, i, a, b):
    return a * i / b - 1 / (2 * b) + a * m / b - round(s * m) / b


def calc_dims_to_pad(m, s):
    epsilon = 1e-8

    if s <= 2:
        return 0

    a = floor(s)
    b = s - a

    if b == 0:
        return 0

    # First case: c_1 <= pi <= c2
    i_lb = get_ilb(m, s, a, b)
    i = ceil(i_lb)
    p_lb = get_plb(m, s, i, a, b)
    p = ceil(max(p_lb, 0))

    if projected_size(m, p, s) == ideal_size(m, s):
        return p

    # Second case: c_1 < pi <= c2
    if abs(round(p_lb) - p_lb) < epsilon:
        p = ceil(max(p_lb + epsilon, 0))

    if projected_size(m, p, s) == ideal_size(m, s):
        return p

    # Third case: c_1 <= pi < c2
    if int(i_lb) == i_lb:
        i = ceil(i_lb + epsilon)
    p_lb = a * i / b - 1 / (2 * b) + a * m / b - round(s * m) / b
    p = ceil(max(p_lb, 0))

    if projected_size(m, p, s) == ideal_size(m, s):
        return p

    # Fourth case: c_1 < pi < c2
    p_lb = a * i / b - 1 / (2 * b) + a * m / b - round(s * m) / b
    p = ceil(max(p_lb, 0))
    if abs(round(p_lb) - p_lb) < epsilon:
        p = ceil(max(p_lb + epsilon, 0))

    if projected_size(m, p, s) == ideal_size(m, s):
        return p


def projected_size(n_slices, p, scale):
    """
    The projected number of slices after initially padding `n_slices`
    by `p`. We would like to choose `p` to match the results of `ideal_slice()`.
    """
    scale_tilde = scale / floor(scale)
    return round((n_slices + p) * scale_tilde) * floor(scale) - round(p * scale)


def ideal_size(n_slices, scale):
    """
    The correct number of slices according to `resize`, which uses
    a `round` operation to get an integer number of slices.
    """
    return round(n_slices * scale)


def find_integer_p(n_slices, s):
    """
    The goal here is to, at test time, pad out the number of slices, then
    run the RCAN model, then crop off all the extra slices we got from the
    initial padding. This function finds the padding which achieves this.
    """
    p = 0  # Start testing from p = 0
    max_iter = 1000  # Maximum number of iterations to prevent infinite loop
    iter_count = 0  # Counter for the number of iterations

    while (
        projected_size(n_slices, p, s) != ideal_size(n_slices, s)
        and iter_count < max_iter
    ):
        p += 1
        iter_count += 1

    # If solution is found within max_iter iterations, return p
    if projected_size(n_slices, p, s) == ideal_size(n_slices, s):
        return p
    # If no solution is found within max_iter iterations, it is unachievable
    # and we just don't do any padding.
    return 0


def calc_dims_to_crop(x, scale):
    return round(x * scale)


def get_patch_edge(sl, patch_size):
    """
    sl: np.array, the slice to draw a patch from
    patch_size: tuple of ints, the patch size in 2D.
    """

    # Choose a random starting position
    sts = [np.random.randint(0, b - p) for b, p in zip(sl.shape, patch_size)]
    ens = [st + p for st, p in zip(sts, patch_size)]
    idx = tuple(slice(st, en) for st, en in zip(sts, ens))

    return sl[idx]


def correct_center_idx(c, edge_minor, edge_major):
    """
    Sometimes, `c` might be too close to the edge. We correct it
    here to lie within the image.
    """
    return min(max(c, edge_minor), edge_major)


def get_patch_center(vol, patch_center, patch_size, return_idx=False):
    """
    vol: np.array, the image volume
    patch_center: tuple of ints, center position of the patch
    patch_size: tuple of ints, the patch size in 3D
    """

    sts = [
        correct_center_idx(c, p // 2, d - p // 2) - p // 2 if p != 1 else c
        for c, p, d in zip(patch_center, patch_size, vol.shape)
    ]
    ens = [st + p for st, p in zip(sts, patch_size)]
    idx = tuple(slice(st, en) for st, en in zip(sts, ens))

    if return_idx:
        return idx

    return vol[idx]


def get_random_centers(vol, patch_size, n_centers, weighted=True):
    if weighted:
        smooth = gaussian_filter(vol, 2.0)
        grads = np.gradient(smooth)
        grad_mag = np.sum([np.sqrt(np.abs(grad)) for grad in grads], axis=0)

        # Set probability to zero at edges
        grad_mag[: patch_size[0] // 2] = 0
        grad_mag[-patch_size[0] // 2 :] = 0

        grad_mag[:, : patch_size[0] // 2] = 0
        grad_mag[:, -patch_size[0] // 2 :] = 0

        # Short-circuit if gradient magnitude is all zero; this
        # happens when the patch size is too big for the dimensions
        # of the image volume
        if grad_mag.sum() == 0:
            print(
                (
                    "Unable to calculate image gradient due to patch"
                    "size and image extents.\n"
                    "Defaulting to uniform patch sampling."
                )
            )
            grad_probs = [None for _ in vol.shape]

        else:
            # Normalize gradient magnitude to create probabilities
            grad_probs = [
                np.array(
                    [
                        np.sum(grad_mag.take(i, axis=axis))
                        for i in range(grad_mag.shape[axis])
                    ]
                )
                for axis in range(grad_mag.ndim)
            ]
            grad_probs = [g / g.sum() for g in grad_probs]

    else:
        grad_probs = [None for _ in vol.shape]

    # Generate random patch centers for each dimension
    centers = [
        np.random.choice(
            np.arange(0, img_dim),
            size=n_centers,
            p=grad_probs[axis],
        )
        for axis, img_dim in enumerate(vol.shape)
    ]
    return centers



import time

import numpy as np
from matplotlib import cycler

try:
    import matplotlib.pyplot as plt
    from IPython import display
    from PIL import Image
except ImportError:
    raise ImportError('Plotting extras must be installed to use plotting featues. '
                      'Install using eclare[plot].')
    
def set_display_mode(mode='dark', figsize=(16, 9), fontsize=18):
    plt.rcParams['figure.figsize'] = figsize
    plt.rcParams['font.size'] = fontsize
    plt.rcParams['image.cmap'] = 'Greys_r'
    plt.rcParams['font.family'] = 'serif'
    plt.rcParams['image.interpolation'] = 'nearest'
    
    if mode == 'dark':
        bg_color = 'black'
        fg_color = 'white'
    else:
        bg_color = 'white'
        fg_color = 'black'
        
    plt.rcParams['figure.facecolor'] = bg_color
    plt.rcParams['axes.facecolor'] = bg_color
    plt.rcParams['axes.edgecolor'] = fg_color
    plt.rcParams['axes.prop_cycle'] = cycler(color=[fg_color]) 
    plt.rcParams['text.color'] = fg_color
    plt.rcParams['axes.labelcolor'] = fg_color
    plt.rcParams['xtick.color'] = fg_color
    plt.rcParams['ytick.color'] = fg_color

def multiplot(imgs, titles, vmins=None, vmaxs=None, figsize=(16, 9), target_shape=None, out_fpath=None, disp=True):
    fig, axs = plt.subplots(
        nrows=1,
        ncols=len(imgs),
        figsize=figsize
    )
    
    if vmins is None:
        vmins = [img.min() for img in imgs]
    if vmaxs is None:
        vmaxs = [img.max() for img in imgs]

    for ax, img, title, vmin, vmax in zip(axs, imgs, titles, vmins, vmaxs):
        if vmin is None:
            vmin = img.min()
        if vmax is None:
            vmax = img.max()
            
        if target_shape is not None:
            # check slice shapes
            if img.shape != (target_shape[0], target_shape[1]):
                # PIL resize needs (y, x)
                img = np.array(Image.fromarray(img)
                             .resize((target_shape[1], target_shape[0]),
                                     Image.NEAREST))
        ax.imshow(np.rot90(img), vmin=vmin, vmax=vmax)
        ax.set_title(title)
        ax.set_xticks([])
        ax.set_yticks([])

    if out_fpath is not None:
        plt.savefig(out_fpath)
    if not disp:
        plt.close()
    else:
        plt.show()


def center_vol_plot(img_vol, target_shape=None, vmin=None, vmax=None, figsize=(16, 9)):
    cs = [c // 2 for c in img_vol.shape]

    x = img_vol[cs[0], :, :]
    y = img_vol[:, cs[1], :]
    z = img_vol[:, :, cs[2]]
    
    if vmin is None:
        vmin = img_vol.min()
    if vmax is None:
        vmax = img_vol.max()

    # For anisotropic images, provide `target_shape` for NN interp
    if target_shape is not None:
        # check slice shapes
        if x.shape != (target_shape[1], target_shape[2]):
            # PIL resize needs (y, x)
            x = np.array(Image.fromarray(x)
                         .resize((target_shape[2], target_shape[1]),
                                 Image.NEAREST))
        if y.shape != (target_shape[0], target_shape[2]):
            y = np.array(Image.fromarray(y)
                         .resize((target_shape[2], target_shape[0]),
                                 Image.NEAREST))
        if z.shape != (target_shape[0], target_shape[1]):
            z = np.array(Image.fromarray(z)
                         .resize((target_shape[1], target_shape[0]),
                                 Image.NEAREST))

    multiplot([x, y, z], [1, 2, 3], 
              vmins=[vmin, vmin, vmin], 
              vmaxs=[vmax, vmax, vmax],
              figsize=figsize,
             )


def anim_paired_patches(lr_patch, hr_patch):
    fig, axs = plt.subplots(1, 2)

    vmin = lr_patch.min()
    vmax = lr_patch.max()

    axs[0].imshow(np.rot90(lr_patch), vmin=vmin, vmax=vmax)
    axs[1].imshow(np.rot90(hr_patch), vmin=vmin, vmax=vmax)

    for ax in axs:
        ax.set_xticks([])
        ax.set_yticks([])

    display.display(plt.show())
    display.clear_output(wait=True)
    time.sleep(.1)



import numpy as np


def fba(imgs, p="infinity"):
    # images to Fourier domain
    vs_hat = [np.fft.rfftn(img) for img in imgs]

    if p == "infinity" or p == "inf":
        out_img = np.max(vs_hat, axis=0)
    else:
        p = float(p)
        # calculate fourier magnitude weights
        denominator = np.sum([np.abs(v_hat) ** p for v_hat in vs_hat], axis=0)
        ws = [np.abs(v_hat) ** p / denominator for v_hat in vs_hat]

        out_img = np.sum([w * v_hat for w, v_hat in zip(ws, vs_hat)], axis=0)

    # return to image domain
    out_img = np.fft.irfftn(out_img).astype(np.float32)

    return out_img



import torch
import numpy as np
from scipy.interpolate import interp1d


def calc_fwhm(kernel):
    """Calculates the full width at half maximum (FWHM) using linear interp.

    Args:
        kernel (numpy.ndarray): The kernel whose FWHM to be calculated.

    Returns
    -------
    fwhm: float
        The calculated FWHM. It is equal to ``right - left``.
    left: float
        The position of the left of the FWHM.
    right: float
        The position of the right of the FWHM.

    """
    kernel = kernel.squeeze()
    half_max = float(np.max(kernel)) / 2
    indices = np.where(kernel > half_max)[0] 
    if len(indices) == 0:
        return 0, 0, 0

    left = indices[0]
    if left > 0:
        interp = interp1d((kernel[left-1], kernel[left]), (left - 1, left))
        left = interp(half_max)
    right = indices[-1]
    if right < len(kernel) - 1:
        interp = interp1d((kernel[right+1], kernel[right]), (right + 1, right))
        right = interp(half_max)
    fwhm = right - left
    return fwhm, left, right


def calc_fwhm_torch(kernel, return_left_right=False):
    """Calculates FWHM using linear interp for pytorch.

    Args:
        kernel (torch.Tensor): The kernel whose FWHM to be calculated.
        return_left_right (bool): Return the left and right coordinates of the
            FWHM if ``True``.

    Returns
    -------
    fwhm: float
        The calculated FWHM. It is equal to ``right - left``.
    left: float
        The position of the left of the FWHM.
    right: float
        The position of the right of the FWHM.

    """
    kernel = kernel.detach().squeeze()
    half_max = kernel.max() / 2
    indices = torch.where(kernel > half_max)[0]

    left = indices[0]
    if left > 0:
        interp = _interp(left - 1, left, kernel[left - 1], kernel[left])
        left  = interp(half_max)

    right = indices[-1]
    if right < len(kernel) - 1:
        interp = _interp(right + 1, right, kernel[right + 1], kernel[right])
        right = interp(half_max)

    fwhm = right - left
    if return_left_right:
        return fwhm, left, right
    else:
        return fwhm


def _interp(x1, x2, y1, y2):
    """1D linear interpolation."""
    a = (x1 - x2) / (y1 - y2)
    b = (x2 * y1 - x1 * y2) / (y1 - y2)
    def func(y):
        return a * y + b
    return func



# Standard imports
import sys
import argparse
import time
from pathlib import Path

# Medical imaging imports
import nibabel as nib
from radifox.utils.resize.affine import update_affine

# ML imports
import numpy as np

# Main function imports
from .eclare import run_eclare
from .espreso import run_espreso

# Utils folder
from .utils.train_set import TrainSet
from .utils.timer import timer_context
from .utils.parse_image_file import parse_image
from .utils.misc_utils import parse_device
from .utils.blur_kernel_ops import parse_kernel

text_div = "=" * 10


def main(args=None):

    main_st = time.time()

    parser = argparse.ArgumentParser()

    parser.add_argument("--in-fpath", type=str, required=True)
    parser.add_argument("--out-dir", type=str, required=True)
    parser.add_argument("--inplane-acq-res", type=float, nargs=2, default=None)
    parser.add_argument("--batch-size", type=int, default=128)
    parser.add_argument("--n-patches", type=int, default=1000000)
    parser.add_argument("--patch-sampling", type=str, default="gradient")
    parser.add_argument("--suffix", type=str, default="_eclare")
    parser.add_argument("--relative-slice-thickness", type=float, default=None)
    parser.add_argument("--relative-slice-profile-type", type=str, default="gaussian")
    parser.add_argument("--relative-slice-profile-fpath", type=Path, default=None)
    parser.add_argument("--gpu-id", type=int, default=-1)
    parser.add_argument("--verbose", action="store_true", default=False)

    args = parser.parse_args(args if args is not None else sys.argv[1:])

    device = parse_device(args.gpu_id)
    inplane_acq_res = args.inplane_acq_res

    in_fpath = Path(args.in_fpath).resolve()
    subj_id = in_fpath.name.split(".nii")[0]
    out_fname = in_fpath.name.replace(".nii", f"{args.suffix}.nii")

    out_dir = Path(args.out_dir).resolve()
    out_dir.mkdir(parents=True, exist_ok=True)
    eclare_out_fpath = out_dir / out_fname

    with timer_context("Parsing image file...", verbose=False):
        image, slice_separation, scales, lr_axis, header, affine, orig_min, orig_max = (
            parse_image(in_fpath, normalize_image=True, inplane_acq_res=inplane_acq_res)
        )

    dataset = TrainSet(image=image, lr_axis=lr_axis, verbose=False)

    # ========== ESPRESO ==========
    if args.relative_slice_thickness is None:
        espreso_psf_fpath = out_dir / f"{subj_id}_espreso_psf.npy"
        espreso_psf_plot_fpath = out_dir / f"{subj_id}_espreso_psf_plot.png"

        g, espreso_elapsed_time = run_espreso(
            slice_separation,
            dataset,
            device,
            espreso_psf_fpath,
            espreso_psf_plot_fpath,
        )

        blur_kernel = g().detach().cpu()
    else:
        blur_kernel = parse_kernel(
            args.relative_slice_profile_fpath,
            args.relative_slice_profile_type,
            args.relative_slice_thickness,
        )

    # ========== ECLARE ==========
    eclare_vol, model_state, eclare_elapsed_time = run_eclare(
        image,
        slice_separation,
        scales,
        lr_axis,
        header,
        affine,
        orig_min,
        orig_max,
        blur_kernel,
        dataset,
        device,
        batch_size=args.batch_size,
        n_patches=args.n_patches,
    )

    # Print final timings and wrap-up.
    if args.relative_slice_thickness is None:
        print(f"\tESPRESO elapsed time: {espreso_elapsed_time:.4f}s\n")
    print(f"\tECLARE elapsed time: {eclare_elapsed_time:.4f}s\n")

    print("Saving image...")
    # Update affine matrix
    scales[lr_axis] = 1 / slice_separation
    new_affine = update_affine(affine, scales)

    # Write nifti
    out_obj = nib.Nifti1Image(eclare_vol, affine=new_affine, header=header)
    nib.save(out_obj, eclare_out_fpath)

    main_en = time.time()

    # Print final timings and wrap-up.
    print(
        (
            "\n\nDONE\n"
            f"Total elapsed time: {main_en - main_st:4f}s\n"
            f"\tWritten to: {eclare_out_fpath}\n"
            f"{text_div} END PREDICTION {text_div}"
        )
    )



"""
This is the generator G for ESPRESO. It learns a PSF and also can apply the PSF.
"""

import torch
import math
from torch import nn
import torch.nn.functional as F

class ConvReLU(nn.Module):
    def __init__(self, in_c, out_c, kernel_size):
        super().__init__()
        self.body = nn.Sequential(
            *[
                nn.Conv2d(in_c, out_c, (kernel_size, 1)),
                nn.ReLU(),
            ]
        )

    def forward(self, x):
        return self.body(x)

class G(nn.Module):    
    def __init__(self, num_channels=256, kernel_size=3, num_convs=2, length=21):
        super().__init__()
        
        size = length + (kernel_size - 1) * num_convs
        self.embedding = nn.Parameter(nn.init.kaiming_uniform_(torch.empty(1, num_channels, size, 1), a=math.sqrt(5)))

        self.body = nn.Sequential(
            *[ConvReLU(num_channels, num_channels, kernel_size) for _ in range(num_convs - 1)],
            nn.Conv2d(num_channels, 1, (kernel_size, 1))
        )

    def forward(self):
        """
        Generate the PSF of shape 1 x 1 x length
        """
        psf = self.body(self.embedding)
        psf = self._symmetrize(psf)
        psf = F.softmax(psf, dim=2)
        
        return psf

    def _symmetrize(self, sp):
        return 0.5 * (sp + torch.flip(sp, (2, )))



import numpy as np
import torch


def rotate_vol_2d(vol, angle):
    """Rotates each 2D slice in a volume by an angle

    Inputs:
        vol: torch Tensor of shape (hr_axis, hr_axis, lr_axis)
        angle: angle to rotate
    Outputs:
        Rotated volume of shape (hr_axis, hr_axis, lr_axis) where rotation occured
        only in the hr-plane.
    """
    # Avoid interpolation errors when able to rot90
    if angle == 0 or angle == 360:
        return vol
    elif angle == 90:
        return torch.rot90(vol, k=1, dims=[0, 1])
    elif angle == -90:
        return torch.rot90(vol, k=-1, dims=[0, 1])
    elif angle == 180:
        return torch.rot90(vol, k=2, dims=[0, 1])
    elif angle == -180:
        return torch.rot90(vol, k=-2, dims=[0, 1])
    elif angle == 270:
        return torch.rot90(vol, k=3, dims=[0, 1])
    elif angle == -270:
        return torch.rot90(vol, k=-3, dims=[0, 1])
    else:
        raise NotImplementedError("Angles other than 90 degree rotations are not supported.")



import numpy as np


def get_pads(target_dim, d):
    if target_dim <= d:
        return 0, 0
    p = (target_dim - d) // 2
    if (p * 2 + d) % 2 != 0:
        return p, p + 1
    return p, p


def target_pad(img, target_dims, mode="zero"):
    pads = tuple(get_pads(t, d) for t, d in zip(target_dims, img.shape))

    if mode == "zero":
        return np.pad(img, pads, mode="constant", constant_values=0), pads
    return np.pad(img, pads, mode=mode), pads


def format_pads(pads):
    """Turn pad amounts into appropriate slices and handle 0 pads as None slices"""
    st = pads[0] if pads[0] != 0 else None
    en = -pads[1] if pads[1] != 0 else None
    return slice(st, en)


def crop(img, pads):
    crops = tuple(map(format_pads, pads))
    return img[crops]



import numpy as np
import nibabel as nib
from radifox.utils.degrade.degrade import fwhm_units_to_voxel_space, fwhm_needed
from radifox.utils.degrade.inplane_res import downsample_k_space


def normalize(x, a=-1, b=1):
    orig_min = x.min()
    orig_max = x.max()

    numer = (x - orig_min) * (b - a)
    denom = orig_max - orig_min

    return a + numer / denom, orig_min, orig_max


def inv_normalize(x, orig_min, orig_max, a=-1, b=1):
    tmp = x - a
    tmp = tmp * (orig_max - orig_min)
    tmp = tmp / (b - a)
    tmp += orig_min
    return tmp


def parse_image(image_file, normalize_image=False, inplane_acq_res=None):
    """
    Open the image volume file, and return pertinent information:
    - The image array as a numpy array
    - The "scale" of the anisotropy (this is the slice separation)
    - The LR axis
    - The header of the image file
    - The affine matrix of the image file
    """
    obj = nib.load(image_file)
    voxel_size = tuple(float(v) for v in obj.header.get_zooms())
    image = obj.get_fdata(dtype=np.float32)
    scales = [1, 1, 1]

    # x, y, and z are the spatial physical measurement sizes
    lr_axis = np.argmax(voxel_size)
    z = voxel_size[lr_axis]
    xy = list(voxel_size)
    xy.remove(z)
    xyz = (xy[0], xy[1], z)
    x, y, z = xyz

    # Exit if the provided image is isotropic through-plane
    assert (
        x != z and y != z
    ), f'Worst resolution found {z} matches one of the better resolutions {x} or {y}; image is "isotropic" and cannot be run through ECLARE.'

    # Correct the in-plane resolution by dropping to the acquired resolution
    if inplane_acq_res is not None:
        print(f"\tReducing in-plane resolutions:")
        print(f"\t({x:.2f}, {y:.2f}) -> ({inplane_acq_res[0]:.2f}, {inplane_acq_res[1]:.2f})")
        scales = [inplane_acq_res[0] / x, inplane_acq_res[1] / y, 1]
        downsample_shape = [int(np.round(d / s)) for d, s in zip(image.shape, scales)]
        image = downsample_k_space(image, target_shape=downsample_shape).astype(np.float32)
        x, y = inplane_acq_res

    slice_separation = round(float(z / min(x, y)), 3)

    if normalize_image:
        image, orig_min, orig_max = normalize(image, 0, 1)
    else:
        orig_min = None
        orig_max = None

    return (
        image,
        slice_separation,
        scales,
        lr_axis,
        obj.header,
        obj.affine,
        orig_min,
        orig_max,
    )


def lr_axis_to_z(img, lr_axis):
    """
    Orient the image volume such that the low-resolution axis
    is in the "z" axis.
    """
    if lr_axis == 0:
        return img.transpose(1, 2, 0)
    elif lr_axis == 1:
        return img.transpose(2, 0, 1)
    elif lr_axis == 2:
        return img


def z_axis_to_lr_axis(img, lr_axis):
    """
    Orient the image volume such that the "z" axis
    is back to the original low-resolution axis
    """
    if lr_axis == 0:
        return img.transpose(2, 0, 1)
    elif lr_axis == 1:
        return img.transpose(1, 2, 0)
    elif lr_axis == 2:
        return img



import numpy as np
import torch
from torch.utils.data import Dataset
from math import ceil
import random

import torch.nn.functional as F
from radifox.utils.resize.pytorch import resize

from .augmentations import augment_3d_image
from .patch_ops import get_patch_edge, get_patch_center, get_random_centers
from .pad import target_pad
from .parse_image_file import lr_axis_to_z, normalize
from .timer import timer_context
from .blur_kernel_ops import calc_extended_patch_size


class TrainSet(Dataset):
    def __init__(self, image, lr_axis, verbose=True):
        self.verbose = verbose
        self.vol = lr_axis_to_z(image, lr_axis)

        # pad the volume to the biggest in-plane square
        with timer_context(
            "Padding image out to square in-plane...", verbose=self.verbose
        ):
            target_shape = (max(self.vol.shape), max(self.vol.shape), self.vol.shape[2])
            self.vol, self.pads = target_pad(
                self.vol,
                target_shape,
            )

    def init_espreso(self, patch_size, n_patches, patch_sampling="gradient"):
        self.algorithm = "ESPRESO"
        self.patch_size = patch_size
        self.throughplane_patch_size = min(self.patch_size)
        self.n_patches = n_patches

        with timer_context(
            "Padding image out to extract patches correctly...", verbose=self.verbose
        ):
            target_shape = [
                s + p
                for s, p in zip(
                    self.vol.shape,
                    (
                        max(self.patch_size),
                        max(self.patch_size),
                        self.throughplane_patch_size,
                    ),
                )
            ]

            # apply the pad
            self.espreso_vol, self.pads = target_pad(
                self.vol,
                target_shape,
                mode="zero",
            )

        with timer_context(
            "Generating (weighted) random patch centers..", verbose=self.verbose
        ):
            self.centers = get_random_centers(
                self.espreso_vol,
                (max(self.patch_size), max(self.patch_size), min(self.patch_size)),
                self.n_patches,
                weighted=patch_sampling == "gradient",
            )

        self.espreso_vol = torch.from_numpy(self.espreso_vol)

    def espreso_getitem(self, i):
        # pull a through-plane patch
        if np.random.uniform() > 0.5:
            throughplane_patch_size = (self.patch_size[0], 1, self.patch_size[1])
        else:
            throughplane_patch_size = (1, self.patch_size[0], self.patch_size[1])

        c = [random.choice(center) for center in self.centers]

        patch_throughplane = get_patch_center(
            self.espreso_vol, c, throughplane_patch_size
        ).squeeze()

        # Data augmentation: random flip along axes
        if np.random.uniform() > 0.5:
            patch_throughplane = torch.flip(patch_throughplane, [0])
        if np.random.uniform() > 0.5:
            patch_throughplane = torch.flip(patch_throughplane, [1])

        # add channel dim and return
        return patch_throughplane.unsqueeze(0)

    def __len__(self):
        return self.n_patches

    def init_eclare(
        self,
        patch_size,
        blur_kernel,
        slice_separation,
        n_patches,
        patch_sampling="gradient",
    ):
        self.algorithm = "ECLARE"
        self.patch_size = patch_size
        self.blur_kernel = blur_kernel
        self.slice_separation = slice_separation
        self.n_patches = n_patches
        self.patch_sampling = patch_sampling

        ext_patch_size, ext_patch_crop = calc_extended_patch_size(
            self.blur_kernel, self.patch_size
        )
        self.patch_size = ext_patch_size
        self.ext_patch_crop = [slice(None, None), slice(None, None), *ext_patch_crop]
        self.n_patches = n_patches

        with timer_context(
            "Padding image out to extract patches correctly...", verbose=self.verbose
        ):
            target_shape = [
                s + p
                for s, p in zip(
                    self.vol.shape, (max(self.patch_size), max(self.patch_size), 0)
                )
            ]

            # apply the pad
            self.eclare_vol, self.pads = target_pad(
                self.vol,
                target_shape,
            )

        self.eclare_vol = torch.from_numpy(self.eclare_vol)

        with timer_context(
            "Generating (weighted) random patch centers..", verbose=self.verbose
        ):
            self.centers = get_random_centers(
                self.eclare_vol,
                (max(self.patch_size), max(self.patch_size), min(self.patch_size)),
                self.n_patches,
                weighted=patch_sampling == "gradient",
            )

    def eclare_getitem(self, i):
        # find the center
        c = [random.choice(center) for center in self.centers]
        # Last index corresponds to the slice. Let's just pull the slice out
        img_slice = self.eclare_vol[:, :, c[-1]]

        # Data augmentation on the whole slice: random flip along first axis
        if np.random.uniform() > 0.5:
            img_slice = torch.flip(img_slice, [0])
            c[0] = img_slice.shape[0] - 1 - c[0]
        # Data augmentation on the whole slice: random flip along second axis
        if np.random.uniform() > 0.5:
            img_slice = torch.flip(img_slice, [1])
            c[1] = img_slice.shape[1] - 1 - c[1]
        # Data augmentation on the whole slice: permute axes
        if np.random.uniform() > 0.5:
            img_slice = torch.permute(img_slice, (1, 0))

        # Pull the extended-size patch
        patch_hr = get_patch_center(img_slice, (c[0], c[1]), self.patch_size)
        patch_hr = patch_hr.unsqueeze(0).unsqueeze(1)

        # Blur the patch
        patch_blur = F.conv2d(patch_hr, self.blur_kernel, padding="same")

        # Crop to true patch size
        patch_hr = patch_hr[self.ext_patch_crop]
        patch_blur = patch_blur[self.ext_patch_crop]

        # Downsample patch blur
        patch_lr = resize(patch_blur, (self.slice_separation, 1), order=3)

        patch_hr = patch_hr.squeeze(0)
        patch_lr = patch_lr.squeeze(0)
        patch_blur = patch_blur.squeeze(0)

        return patch_lr, patch_hr

    def __getitem__(self, i):
        if self.algorithm == "ESPRESO":
            return self.espreso_getitem(i)
        elif self.algorithm == "ECLARE":
            return self.eclare_getitem(i)



import numpy as np
import torch
from radifox.utils.degrade.degrade import select_kernel
from math import ceil


def parse_kernel(blur_kernel_file, blur_kernel_type, blur_fwhm):

    if blur_kernel_file is not None:
        blur_kernel = np.load(blur_kernel_file)
    else:
        window_size = int(2 * round(blur_fwhm) + 1)
        blur_kernel = select_kernel(window_size, blur_kernel_type, fwhm=blur_fwhm)
    blur_kernel /= blur_kernel.sum()
    blur_kernel = blur_kernel.squeeze()[None, None, :, None]
    blur_kernel = torch.from_numpy(blur_kernel).float()

    return blur_kernel


def calc_extended_patch_size(blur_kernel, patch_size):
    """
    Calculate the extended patch size. This is necessary to remove all boundary
    effects which could occur when we apply the blur kernel. We will pull a patch
    which is the specified patch size plus half the size of the blur kernel. Then we later
    blur at test time, crop off this extended patch size, then downsample.
    """

    L = blur_kernel.shape[0]

    ext_patch_size = [p + 2 * ceil(L / 2) if p != 1 else p for p in patch_size]
    ext_patch_crop = [(e - p) // 2 for e, p in zip(ext_patch_size, patch_size)]
    ext_patch_crop = tuple([slice(d, -d) for d in ext_patch_crop if d != 0])

    return ext_patch_size, ext_patch_crop



import torch
import numpy as np
import torch.nn.functional as F


def unwrap(x):
    pi = torch.pi
    y = x % (2 * pi)
    return torch.where(y > pi, 2 * pi - y, y)


def power_of_2_pad(x):
    """
    In order to compute FFT in half-precision, all dimensions
    must be powers of 2. This function takes `x` and returns
    a reflect-padded `x_pad` where all axis are powers of 2.
    """
    # find the next power of two
    target_shape = [
        2 ** ((d - 1).bit_length()) if i > 1 else 0 for i, d in enumerate(x.shape)
    ]

    # pads
    padding = [new - old for new, old in zip(target_shape, x.shape)]

    x_pad = F.pad(x, (0, padding[-2], 0, padding[-1]), mode="reflect")
    return x_pad


def mag_pha(x, log=False):
    omega = torch.fft.rfftn(power_of_2_pad(x))

    mag = torch.abs(omega).float()
    if log:
        mag = torch.log(mag + 1e-8)

    pha = torch.angle(omega).float()
    pha = unwrap(pha)

    return mag, pha


class GANLoss(torch.nn.Module):
    r"""Loss of the original GAN with cross entropy.

    For the discriminator :math:`D`, this loss minimizes the binary cross
    entropy with logits:

    .. math::

        l = - \mathrm{mean}_x (y \ln(\sigma(D(x)))
          + (1 - y) \ln(1 - \sigma(D(x)))),

    where :math:`\sigma` is the sigmoid function. If
    :math:`x \in \mathrm{\{truth\}}`, we have :math:`y = 1` and

    .. math::

        l = - \mathrm{mean}_x \ln(\sigma(D(x))).

    If :math:`x \in \mathrm{\{generated\}}`, i.e. :math:`\exists z` s.t.
    :math:`x = G(z)` where `G` is the generator, we have :math:`y = 0` and

    .. math::

        l = - \mathrm{mean}_x \ln(1 - \sigma(D(G(z)))).

    Combine the above two terms together, we can get the loss for the
    discriminator.

    For the generator :math:`G`, this loss minimizes the binary cross entropy
    with the same form and :math:`y = 1`, which is to minimize

    .. math::

        l = - \mathrm{mean}_x \ln(\sigma(D(G(z)))).

    This is the modified GAN loss which minimizes
    :math:`l = \mathrm{mean}_x \ln(1 - \sigma(D(G(z))))`.

    """

    def __init__(self):
        super().__init__()

    def forward(self, x, is_real):
        """Calculates the loss.

        Args:
            x (torch.Tensor): The image.
            is_real (bool): Whether the image is real.

        Returns:
            The calculated loss.

        """
        target = torch.ones_like(x) - 0.1 if is_real else torch.zeros_like(x) + 0.1
        loss = F.binary_cross_entropy_with_logits(x, target)
        return loss


class SmoothnessLoss(torch.nn.Module):
    r"""L2 norm of derivative.

    This loss minimizes

    .. math::

        l = \lVert \nabla k \rVert_2,

    where :math:`k` is the kernel, to encourage smoothness.

    """

    def forward(self, kernel):
        """Calculates the loss for this kernel."""
        device = kernel.device
        operator = torch.tensor([1, -1], dtype=torch.float32, device=device)
        operator = operator[None, None, ..., None]
        derivative = F.conv2d(kernel, operator)
        loss = torch.sqrt(torch.sum(derivative**2))
        return loss


class CenterLoss(torch.nn.Module):
    r"""Penalizes off-center.

    This loss minimizes the differences between the center of the kernel and the
    center of the vector:

    .. math::

        l = \left(\sum_x k(x) x - C \right) ^ 2,

    where :math:`k(x)` is the kernel, :math:`x` is vector indices, and :math:`C`
    is the center of the vector. Assume the sum of the kernel equals 1.

    """

    def __init__(self, kernel_length):
        super().__init__()
        self.kernel_length = kernel_length
        center = torch.tensor(self.kernel_length // 2, dtype=torch.float32)
        self.register_buffer("center", center)
        locs = torch.arange(self.kernel_length, dtype=torch.float32)
        self.register_buffer("locs", locs)

    def forward(self, kernel):
        """Calculates the loss for this kernel."""
        kernel_center = torch.sum(kernel.squeeze() * self.locs)
        loss = F.mse_loss(kernel_center, self.center)
        return loss


class BoundaryLoss(torch.nn.Module):
    r"""Penalizes non-zero values at kernel boundary.

    This loss minimizes the weighted sum:

    .. math::

        l = \sum_x | m(x) k(x) |

    where :math:`m` is 0, 0, 1, 1, ..., 1, 1, 0, 0.

    """

    def __init__(self, kernel_length):
        super().__init__()
        self.kernel_length = kernel_length
        mask = self._create_penalty_mask().clone().detach().float()
        self.register_buffer("mask", mask[None, None, ..., None])

    def _create_penalty_mask(self):
        mask = torch.ones(self.kernel_length).float()
        mask[2:-2] = 0
        return mask

    def forward(self, kernel):
        """Calculates the loss for this kernel."""
        return torch.sum(torch.abs(kernel * self.mask))


class PeakLoss(torch.nn.Module):
    def forward(self, kernel):
        device = kernel.device
        op_f = torch.tensor([1, -1], dtype=torch.float32, device=device)
        op_b = torch.tensor([-1, 1], dtype=torch.float32, device=device)
        op_f = op_f[None, None, ..., None]
        op_b = op_b[None, None, ..., None]
        diff_f = F.conv2d(F.pad(kernel, (0, 0, 0, 1)), op_f).squeeze()
        diff_b = F.conv2d(F.pad(kernel, (0, 0, 1, 0)), op_b).squeeze()
        mid_ind = kernel.shape[2] // 2
        result = torch.sum(F.relu(diff_f[:mid_ind])) + torch.sum(
            F.relu(-diff_f[mid_ind:])
        )
        # result = torch.tensor(0).float().cuda()
        # diff = F.relu(diff_f * diff_b).squeeze()
        # result = torch.sum(diff[:mid_ind]) + torch.sum(diff[mid_ind+1:])
        return result



"""
This is the discriminator D for ESPRESO.

TODO: recode
"""

import torch
import math
from torch import nn
import torch.nn.utils.spectral_norm as sn
import torch.nn.functional as F

class ConvSpectnormLReLU(nn.Module):
    def __init__(self, in_c, out_c, kernel_size, alpha):
        super().__init__()
        self.body = nn.Sequential(
            *[
                sn(nn.Conv2d(in_c, out_c, kernel_size)),
                nn.LeakyReLU(alpha),                
            ]
        )

    def forward(self, x):
        return self.body(x)

class D(nn.Module):
    def __init__(self, 
                 inp_ch=1,
                 out_chs=(64, 64, 64, 64),
                 kernel_sizes=((3, 1), (3, 1), (3, 1), (3, 1), (3, 1)),
                 alpha=0.1):

        
        super().__init__()
        assert len(out_chs) == len(kernel_sizes) - 1

        inp_cs = [inp_ch, *out_chs[:-1]]
        out_cs = out_chs[1:]

        self.body = nn.Sequential(
            *[
                ConvSpectnormLReLU(in_c, out_c, kernel_size, alpha)
                for in_c, out_c, kernel_size in zip(inp_cs, out_cs, kernel_sizes)
            ],
            sn(nn.Conv2d(out_chs[-1], 1, kernel_sizes[-1])),
        )
            
    def forward(self, x):
        return self.body(x)